<?php

namespace jrosset\SemVer;

use DirectoryIterator;
use Exception;
use InvalidArgumentException;
use SplFileInfo;

/**
 * A semantic version
 *
 * @link https://semver.org/ The semantic versionning (2.0.0)
 */
final class Version {
    /**
     * The separator between version number
     */
    private const SEPARATOR_VERSION = '.';
    /**
     * The separator for pre-release
     */
    private const SEPARATOR_PRE_RELEASE = '-';
    /**
     * The separator for build metadata
     */
    private const SEPARATOR_BUILD_METADATA = '+';

    /**
     * @var string The regular expression's pattern for the semantic version parsing
     *
     * @link https://regex101.com/r/gDKcBH/1
     */
    private static string $constRegexPattern = '';

    /**
     * @var int The major version
     */
    private int $major;
    /**
     * @var int The minor version
     */
    private int $minor;
    /**
     * @var int The patch version
     */
    private int $patch;
    /**
     * @var VersionExtraPart The pre-release
     */
    private VersionExtraPart $preRelease;
    /**
     * @var VersionExtraPart The build metadata
     */
    private VersionExtraPart $buildMetadata;

    /**
     * Create a semantic version
     *
     * @param int|null                     $major         The major version
     * @param int|null                     $minor         The minor version
     * @param int|null                     $patch         The patch version
     * @param VersionExtraPart|string|null $preRelease    The pre-release
     * @param VersionExtraPart|string|null $buildMetadata The build metadata
     *
     * @throws InvalidArgumentException If the semantic version is not valid
     */
    public function __construct (?int $major = null, ?int $minor = null, ?int $patch = null, $preRelease = null, $buildMetadata = null) {
        $this->setMajor($major);
        $this->setMinor($minor);
        $this->setPatch($patch);

        if (!$this->isEmpty()) {
            $this->setPreRelease($preRelease);
            $this->setBuildMetadata($buildMetadata);
        }
    }
    /**
     * Export the semantic version to a string
     *
     * @return string The semantic version to a string
     */
    public function __toString (): string {
        if ($this->isEmpty()) {
            return '';
        }

        $string = $this->getMajor() . self::SEPARATOR_VERSION . $this->getMinor() . self::SEPARATOR_VERSION . $this->getPatch();
        if (!$this->getPreRelease()->isNull()) {
            $string .= self::SEPARATOR_PRE_RELEASE . $this->getPreRelease();
        }
        if (!$this->getBuildMetadata()->isNull()) {
            $string .= self::SEPARATOR_BUILD_METADATA . $this->getBuildMetadata();
        }

        return $string;
    }
    /**
     * Export the semantic version to a <u>short</u> string (only version numbers)
     *
     * @param bool $showEmptyPatch Show the patch version if null/0 ?
     *
     * @return string The semantic version to a <u>short</u> string
     */
    public function toShortString (bool $showEmptyPatch = false): string {
        if ($this->isEmpty()) {
            return '';
        }

        return $this->getMajor() . self::SEPARATOR_VERSION . $this->getMinor() . (!$showEmptyPatch && $this->getPatch() === 0 ? '' : self::SEPARATOR_VERSION . $this->getPatch());
    }

    /**
     * Parse a semantic version from a string
     *
     * @param string|null $semanticVersion The string to parse
     *
     * @return self The corresponding semantic version
     *
     * @throws InvalidArgumentException If the semantic version is not valid
     */
    public static function parse (?string $semanticVersion): self {
        //region Construction pattern regex
        if (self::$constRegexPattern === '') {
            /** @noinspection RegExpSuspiciousBackref */
            self::$constRegexPattern = /** @lang PhpRegExp */
                '/'
                . '(?(DEFINE)(?\'VersionNumber\'0|[1-9][0-9]*))'
                . '(?(DEFINE)(?\'Identifier\'' . VersionExtraPart::REGEX_IDENTIFIER . '))'
                . '(?(DEFINE)(?\'ExtraPart\'(?P>Identifier)(?:' . preg_quote(VersionExtraPart::SEPARATOR_IDENTIFIERS, '/') . '(?P>Identifier))*))'
                . '^v?'
                . '(?<major>(?P>VersionNumber))(?:' . preg_quote(self::SEPARATOR_VERSION, '/')
                . '(?<minor>(?P>VersionNumber))(?:' . preg_quote(self::SEPARATOR_VERSION, '/')
                . '(?<patch>(?P>VersionNumber)))?)?'
                . '(?:' . preg_quote(self::SEPARATOR_PRE_RELEASE, '/') . '(?<preRelease>(?P>ExtraPart)))?'
                . '(?:' . preg_quote(self::SEPARATOR_BUILD_METADATA, '/') . '(?<buildMetadata>(?P>ExtraPart)))?'
                . '$/';
        }
        //endregion

        if (preg_match(self::$constRegexPattern, $semanticVersion, $match) !== 1) {
            $match = [];
        }
        return new self (
            $match['major'] ?? 0,
            $match['minor'] ?? 0,
            $match['patch'] ?? 0,
            $match['preRelease'] ?? new VersionExtraPart(),
            $match['buildMetadata'] ?? new VersionExtraPart(),
        );
    }

    /**
     * Compare with another semantic version
     *
     * @param self $other The other semantic version
     *
     * @return int 1 if current version is greater, 0 if they are equal, -1 if current version is lower
     */
    public function compare (self $other): int {
        if ($this->getMajor() > $other->getMajor()) {
            return 1;
        }
        elseif ($this->getMajor() < $other->getMajor()) {
            return -1;
        }

        if ($this->getMinor() > $other->getMinor()) {
            return 1;
        }
        elseif ($this->getMinor() < $other->getMinor()) {
            return -1;
        }

        if ($this->getPatch() > $other->getPatch()) {
            return 1;
        }
        elseif ($this->getPatch() < $other->getPatch()) {
            return -1;
        }

        if (($compare = $this->getPreRelease()->compare($other->getPreRelease())) !== 0) {
            return $compare;
        }
        if (($compare = $this->getBuildMetadata()->compare($other->getBuildMetadata())) !== 0) {
            return $compare;
        }

        return 0;
    }

    /**
     * Is the semantic version empty ?
     *
     * @return bool Is the semantic version empty ?
     */
    public function isEmpty (): bool {
        return $this->getMajor() === 0
               && $this->getMinor() === 0
               && $this->getPatch() === 0;
    }
    /**
     * Alias for {@see Version::isNull()}
     */
    public function isNull (): bool {
        return $this->isEmpty();
    }

    /**
     * The major version
     *
     * @param bool $asNull Return null if 0
     *
     * @return int|null The major version
     */
    public function getMajor (bool $asNull = false): ?int {
        return $asNull && $this->major === 0 ? null : $this->major;
    }
    /**
     * Set the major version
     *
     * @param int|null $major The new major version
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the major version is not valid
     */
    public function setMajor (?int $major): self {
        if ($major < 0) {
            throw new InvalidArgumentException('The major version is not valid');
        }

        $this->major = $major ?? 0;
        return $this;
    }

    /**
     * The minor version
     *
     * @param bool $asNull Return null if 0
     *
     * @return int|null The minor version
     */
    public function getMinor (bool $asNull = false): ?int {
        return $asNull && $this->minor === 0 ? null : $this->minor;
    }
    /**
     * Set the minor version
     *
     * @param int|null $minor The new minor version
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the minor version is not valid
     */
    public function setMinor (?int $minor): self {
        if ($minor < 0) {
            throw new InvalidArgumentException('The minor version is not valid');
        }

        $this->minor = $minor ?? 0;
        return $this;
    }

    /**
     * The patch version
     *
     * @param bool $asNull Return null if 0
     *
     * @return int|null The patch version
     */
    public function getPatch (bool $asNull = false): ?int {
        return $asNull && $this->patch === 0 ? null : $this->patch;
    }
    /**
     * Set the patch version
     *
     * @param int|null $patch The new patch version
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the path version is not valid
     */
    public function setPatch (?int $patch): self {
        if ($patch < 0) {
            throw new InvalidArgumentException('The path version is not valid');
        }

        $this->patch = $patch ?? 0;
        return $this;
    }

    /**
     * The pre-release
     *
     * @param bool $asNull Return null if no pre-release
     *
     * @return VersionExtraPart|null The pre-release
     */
    public function getPreRelease (bool $asNull = false): ?VersionExtraPart {
        return $asNull && $this->preRelease->isNull() ? null : $this->preRelease;
    }
    /**
     * Set the pre-release
     *
     * @param VersionExtraPart|string|null $preRelease The new pre-release
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the pre-release is not valid
     */
    public function setPreRelease ($preRelease): self {
        try {
            $this->preRelease = is_string($preRelease)
                ? VersionExtraPart::parse($preRelease)
                : $preRelease ?? new VersionExtraPart();
        }
        catch (Exception $exception) {
            throw new InvalidArgumentException('The pre-release is not valid', 0, $exception);
        }

        return $this;
    }

    /**
     * The build metadata
     *
     * @param bool $asNull Return null if no build metadata
     *
     * @return VersionExtraPart|null The build metadata
     */
    public function getBuildMetadata (bool $asNull = false): ?VersionExtraPart {
        return $asNull && $this->buildMetadata->isNull() ? null : $this->buildMetadata;
    }
    /**
     * Set the build metadata
     *
     * @param VersionExtraPart|string|null $buildMetadata The new build metadata
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the build metadata is not valid
     */
    public function setBuildMetadata ($buildMetadata): self {
        try {
            $this->buildMetadata = is_string($buildMetadata)
                ? VersionExtraPart::parse($buildMetadata)
                : $buildMetadata ?? new VersionExtraPart();
        }
        catch (Exception $exception) {
            throw new InvalidArgumentException('The pre-release is not valid', 0, $exception);
        }
        return $this;
    }

    //region Extraction depuis Git
    /**
     * Subdirectory of git for tags
     */
    protected const PATH_GIT_TAGS = '/.git/refs/tags';

    /**
     * Get last version from git (scan in tags)
     *
     * @param string $repositoryRootDir The git repository root directory (without the ".git")
     *
     * @return self The last version or an empty semantic version if none
     */
    public static function getGitLastTag (string $repositoryRootDir): self {
        $versions = [];

        /** @var SplFileInfo $file */
        $directoryIterator = new DirectoryIterator($repositoryRootDir . self::PATH_GIT_TAGS);
        foreach ($directoryIterator as $file) {
            if (!$file->isFile()) {
                continue;
            }
            if (($version = self::parse($file->getFilename()))->isEmpty()) {
                continue;
            }
            $versions[] = $version;
        }
        if (count($versions) === 0) {
            return new self();
        }

        usort(
            $versions,
            function (self $versionNumber1, self $versionNumber2): int {
                return $versionNumber1->compare($versionNumber2);
            }
        );
        return array_pop($versions);
    }
    //endregion
}