<?php

namespace jrosset\SemVer;

use InvalidArgumentException;

/**
 * An extra part to version : pre-release or build metadata
 */
class VersionExtraPart {
    /**
     * The separator between identifiers
     */
    public const SEPARATOR_IDENTIFIERS = '.';
    /**
     * The regular expression's pattern for a valid identifier
     */
    public const REGEX_IDENTIFIER = '[1-9a-zA-Z-][0-9a-zA-Z-]*';
    /**
     * The regular expression's pattern for a valid identifier
     */
    public const REGEX_IDENTIFIER_PHP = /** @lang PhpRegExp */
        '/^' . self::REGEX_IDENTIFIER . '$/';

    /**
     * @var string[] The identifiers list
     */
    private array $identifiers;

    /**
     * Create an extra part
     *
     * @param string[]|null $identifiers The identifiers list
     *
     * @throws InvalidArgumentException If one of the identifier is not valid (cf. {@see VersionExtraPart::REGEX_IDENTIFIER_PHP})
     */
    public function __construct (?array $identifiers = null) {
        $this->setIdentifiers($identifiers);
    }
    /**
     * Export the extra part to a string
     *
     * @return string The extra part to a string
     */
    public function __toString (): string {
        return implode(self::SEPARATOR_IDENTIFIERS, $this->getIdentifiers());
    }

    /**
     * Parse an extra part from a string
     *
     * @param string|null $extraPart The string to parse
     *
     * @return self The corresponding extra part
     *
     * @throws InvalidArgumentException If one of the identifier is not valid (cf. {@see VersionExtraPart::REGEX_IDENTIFIER_PHP})
     */
    public static function parse (?string $extraPart): self {
        return new self(explode(self::SEPARATOR_IDENTIFIERS, $extraPart ?? ''));
    }

    /**
     * Is the extra part empty ?
     *
     * @return bool Is the extra part empty ?
     */
    public function isEmpty (): bool {
        return count($this->identifiers) === 0;
    }
    /**
     * Alias for {@see VersionExtraPart::isNull()}
     */
    public function isNull (): bool {
        return $this->isEmpty();
    }

    /**
     * Compare with another extra part
     *
     * @param VersionExtraPart $other The other extra part
     *
     * @return int 1 if greater than, 0 is equal or -1 if lesser than
     */
    public function compare (VersionExtraPart $other): int {
        $identifiers1 = $this->getIdentifiers();
        $identifiers2 = $other->getIdentifiers();

        $maxSize = max(count($identifiers1), count($identifiers2));
        $identifiers1 = array_pad($identifiers1, $maxSize, null);
        $identifiers2 = array_pad($identifiers2, $maxSize, null);

        for ($iterator = 0; $iterator < $maxSize; $iterator++) {
            if (($compare = self::compareIdentifier($identifiers1[$iterator], $identifiers2[$iterator])) === 0) {
                continue;
            }

            return $compare;
        }

        return 0;
    }
    /**
     * Compare one identifier with another one
     *
     * @param string|null $identifier1 The first identifier
     * @param string|null $identifier2 The second identifier
     *
     * @return int 1 if the first is greater than the second, 0 is they are equals or -1 if the first is lesser than the second
     */
    protected static function compareIdentifier (?string $identifier1, ?string $identifier2): int {
        //region Check set size (null if missing)
        if ($identifier1 === null && $identifier2 === null) {
            return 0;
        }
        elseif ($identifier1 !== null && $identifier2 === null) {
            return 1;
        }
        elseif ($identifier1 === null && $identifier2 !== null) {
            return -1;
        }
        //endregion
        //region Check digit-only vs string
        if (!self::isOnlyDigits($identifier1) && self::isOnlyDigits($identifier2)) {
            return 1;
        }
        elseif (self::isOnlyDigits($identifier1) && !self::isOnlyDigits($identifier2)) {
            return -1;
        }
        //endregion
        //region The comparison it-self
        if ($identifier1 == $identifier2) {
            return 0;
        }
        else {
            return $identifier1 > $identifier2 ? 1 : -1;
        }
        //endregion
    }
    /**
     * Check if a string contains only digits
     *
     * @param string $string The string to check
     *
     * @return bool Is the string contains only digits ?
     */
    protected static function isOnlyDigits (string $string): bool {
        return preg_match('/^[0-9]+$/', $string) === 1;
    }

    /**
     * The identifiers list
     *
     * @param bool $asNull Return null if empty
     *
     * @return string[]|null The identifiers list
     */
    public function getIdentifiers (bool $asNull = false): ?array {
        return $asNull && $this->isNull() ? null : $this->identifiers;
    }
    /**
     * Set the identifiers list
     *
     * @param array|null $identifiers The new identifiers list
     *
     * @return $this
     *
     * @throws InvalidArgumentException If one of the identifier is not valid (cf. {@see VersionExtraPart::REGEX_IDENTIFIER_PHP})
     */
    protected function setIdentifiers (?array $identifiers): self {
        if ($identifiers !== null) {
            foreach ($identifiers as $identifier) {
                if (preg_match(self::REGEX_IDENTIFIER_PHP, $identifier) !== 1) {
                    throw new InvalidArgumentException('The identifiant is not valid: ' . $identifier);
                }
            }
        }

        $this->identifiers = $identifiers ?? [];
        return $this;
    }
}